/*
 * $Id:$
 *
 * Copyright (C) 2004 Essence project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <Edb.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

#include "essence_db.h"
#include "essence_db_item.h"
#include "e-cell-renderer-combo.h"
#include "misc.h"
#include "../pixmaps/close.h"

static void on_text_edited (GtkCellRendererText *renderer,
                            gchar *path, gchar *value, EssenceDb *edb)
{
	EssenceDbItem *ei = NULL;
	GtkTreeModel *model = GTK_TREE_MODEL (edb->store);
	GtkTreeIter iter;
	int col, ival;
	float fval;
	gchar *type, buf[128];

	gtk_tree_selection_get_selected (edb->select, &model, &iter);
	col = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (renderer),
	                       "col"));

	/* get the EssenceDbItem that's associated with this entry */
	gtk_tree_model_get (model, &iter, COL_ITEM, &ei, -1);
	assert (ei);

	if (col == COL_KEY) {
		gtk_list_store_set (edb->store, &iter, col, value, -1);
		essence_db_item_key_set (ei, value);
		return;
	}

	if (col == COL_TYPE) {
		gtk_list_store_set (edb->store, &iter, col, value, -1);

		/* if the new data type is int or float, we'll have to make
		 * sure the value column doesn't have invalid data.
		 */
		switch (*value) {
			case 'S':
				ei->type = ITEM_TYPE_STR;
				return;
				break;
			case 'I':
				ei->type = ITEM_TYPE_INT;
				break;
			case 'F':
				ei->type = ITEM_TYPE_FLOAT;
				break;
			default:
				assert (false);
				break;
		}

		gtk_tree_model_get (model, &iter, COL_VALUE, &value, -1);
		col = COL_VALUE;
	}

	/* the value column has been edited, so we need to find out
	 * which data type is to be used.
	 * The possible values are "String", "Float" and "Integer"
	 */
	gtk_tree_model_get (model, &iter, COL_TYPE, &type, -1);

	switch (*type) {
		case 'S':
			gtk_list_store_set (edb->store, &iter, col, value, -1);
			strlcpy (ei->data.s, value, sizeof (ei->data.s));
			break;
		case 'I':
			ival = atoi (value);
			snprintf (buf, sizeof (buf), "%i", ival);
			gtk_list_store_set (edb->store, &iter, col, buf, -1);
			ei->data.i = ival;
			break;
		case 'F':
			fval = atof (value);
			snprintf (buf, sizeof (buf), "%f", fval);
			gtk_list_store_set (edb->store, &iter, col, buf, -1);
			ei->data.f = fval;
			break;
		default:
			assert (false);
			break;
	}

	g_free (type);
}

/**
 * Creates a GtkListStore and GtkTreeView for the EDB items.
 */
static GtkTreeView *create_tree (EssenceDb *edb)
{
	GtkTreeView *tree;
	GtkTreeModel *model;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *col;
	GList *list = NULL;
	const char *caption[COL_NUM] = {"Key", "Type", "Value", ""};
	int i;

	edb->store = gtk_list_store_new (COL_NUM, G_TYPE_STRING,
	                                 G_TYPE_STRING, G_TYPE_STRING,
	                                 G_TYPE_POINTER);

	model = GTK_TREE_MODEL (edb->store);
	tree = GTK_TREE_VIEW (gtk_tree_view_new_with_model (model));
	g_object_unref (edb->store);

	/* create a list that contains the possible values for the
	 * "type" column.
	 */
	list = g_list_prepend (list, "String");
	list = g_list_prepend (list, "Float");
	list = g_list_prepend (list, "Integer");

	/* create columns */
	for (i = 0; i < COL_NUM; i++) {
		if (i == COL_TYPE) {
			renderer = e_cell_renderer_combo_new ();
			g_object_set (G_OBJECT (renderer), "list", list, NULL);
		} else
			renderer = gtk_cell_renderer_text_new ();

		g_object_set_data (G_OBJECT (renderer), "col",
		                   GINT_TO_POINTER (i));
		g_object_set (G_OBJECT (renderer), "editable", TRUE, NULL);

		g_signal_connect (G_OBJECT (renderer), "edited",
		                  G_CALLBACK (on_text_edited), edb);

		col = gtk_tree_view_column_new_with_attributes (caption[i],
		                                                renderer, NULL);

		if (i != COL_ITEM)
			gtk_tree_view_column_add_attribute (col, renderer, "text", i);

		if (i == COL_TYPE) {
			gtk_tree_view_column_set_fixed_width (col, 75);
			gtk_tree_view_column_set_sizing (col,
			                                 GTK_TREE_VIEW_COLUMN_FIXED);
		}

		gtk_tree_view_column_set_sort_column_id (col, i);
		gtk_tree_view_column_set_resizable (col, TRUE);
		gtk_tree_view_column_set_reorderable (col, TRUE);
		gtk_tree_view_column_set_visible (col, i != COL_ITEM);
		gtk_tree_view_append_column (tree, col);

		edb->select = gtk_tree_view_get_selection (tree);
		gtk_tree_selection_set_mode (edb->select, GTK_SELECTION_SINGLE);
	}

	/* sort by key */
	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (model),
	                                      COL_KEY, GTK_SORT_ASCENDING);

	return tree;
}

/**
 * Add a new row to the treeview/liststore.
 */
static void on_add_button_clicked (GtkButton *b, EssenceDb *edb)
{
	EssenceDbItem *ei;
	GtkTreeModel *model;
	GtkTreeIter iter;
	GtkTreePath *path;
	GtkTreeViewColumn *col;

	if (!(ei = essence_db_item_new (edb->db)))
		return;

	ei->type = ITEM_TYPE_INT;
	essence_db_item_key_set (ei, "New Entry");

	edb->items = g_slist_append (edb->items, ei);
	iter = essence_db_item_show (ei, edb->store);

	/* now scroll to the new row */
	model = GTK_TREE_MODEL (edb->store);
	path = gtk_tree_model_get_path (model, &iter);
	col = gtk_tree_view_get_column (edb->tree, COL_KEY);

	/* start editing the key */
	gtk_tree_view_set_cursor (edb->tree, path, col, TRUE);

	gtk_tree_path_free (path);
}

static void on_remove_button_clicked (GtkButton *b, EssenceDb *edb)
{
	EssenceDbItem *ei = NULL;
	GtkTreeModel *model = GTK_TREE_MODEL (edb->store);
	GtkTreeIter iter;

	if (!gtk_tree_selection_get_selected (edb->select, &model,
	                                      &iter))
		return;

	/* get the EssenceDbItem that's associated with this entry */
	gtk_tree_model_get (model, &iter, COL_ITEM, &ei, -1);
	assert (ei);

	/* remove the row from the liststore */
	gtk_list_store_remove (edb->store, &iter);

	/* remove the EssenceDbItem from our list and from the db */
	essence_db_item_remove (ei);

	edb->items = g_slist_remove (edb->items, ei);
	essence_db_item_free (ei);
}

/**
 * Creates a button group which contains the "Add" and "Remove" buttons.
 */
static GtkWidget *create_button_group (EssenceDb *edb)
{
	GtkWidget *bbox, *button;

	bbox = gtk_hbutton_box_new ();

	gtk_container_set_border_width (GTK_CONTAINER (bbox), 12);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_END);
	gtk_box_set_spacing (GTK_BOX (bbox), 6);

	/* add "Add" button */
	button = gtk_button_new_from_stock (GTK_STOCK_ADD);
	gtk_container_add (GTK_CONTAINER (bbox), button);

	g_signal_connect (G_OBJECT (button), "clicked",
	                  G_CALLBACK (on_add_button_clicked), edb);

	/* add "Remove" button */
	button = gtk_button_new_from_stock (GTK_STOCK_REMOVE);
	gtk_container_add (GTK_CONTAINER (bbox), button);

	g_signal_connect (G_OBJECT (button), "clicked",
	                  G_CALLBACK (on_remove_button_clicked), edb);

	return bbox;
}

static void on_tab_close_button_clicked (GtkButton *b, EssenceDb *edb)
{
	edb->dest_notify (edb->dest_notify_data, edb);
	essence_db_free (edb);
}

static GtkWidget *create_notebook_page_header (EssenceDb *edb,
                                               const char *caption,
	                                           GtkWidget **label)
{
	GtkWidget *hbox, *image, *button;
	char markup[PATH_MAX + 1];
	static GdkPixbuf *pic = NULL;

	assert (edb);
	assert (caption);

	hbox = gtk_hbox_new (FALSE, 0);

	/* create the label */
	snprintf (markup, sizeof (markup),
	          "<span weight=\"bold\">%s</span>", caption);
	*label = gtk_label_new (markup);
	gtk_label_set_use_markup (GTK_LABEL (*label), TRUE);
	gtk_box_pack_start (GTK_BOX (hbox), *label,
	                    FALSE, FALSE, 3);

	/* create the button */
	button = gtk_button_new ();
	gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);

	g_signal_connect (G_OBJECT (button), "clicked",
	                  G_CALLBACK (on_tab_close_button_clicked), edb);
	gtk_box_pack_start (GTK_BOX(hbox), button, FALSE, FALSE, 0);

	/* put the close icon in the button */
	if (!pic)
		pic = gdk_pixbuf_new_from_inline (-1, pic_close, FALSE, NULL);

	image = gtk_image_new_from_pixbuf (pic);

	gtk_container_add (GTK_CONTAINER (button), image);

	gtk_widget_show_all (hbox);

	return hbox;
}

EssenceDb *essence_db_new (GtkNotebook *notebook, const char *file,
                           GFunc dest_notify, gpointer udata)
{
	EssenceDb *edb;
	GtkWidget *scrolled, *vbox, *tab_header;
	char *name;

	assert (notebook);

	if (!file_exists (file))
		return NULL;

	if (!(edb = calloc (1, sizeof (EssenceDb))))
		return NULL;

	edb->dest_notify = dest_notify;
	edb->dest_notify_data = udata;
	edb->notebook = notebook;
	strlcpy (edb->file, file, sizeof (edb->file));

	name = g_path_get_basename ((char *) file);

	/* create a vbox for the tree and the button box */
	vbox = gtk_vbox_new (FALSE, 6);

	/* create a scrolled window */
	scrolled = gtk_scrolled_window_new (NULL, NULL);
	gtk_container_set_border_width (GTK_CONTAINER (scrolled), 12);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled),
	                                GTK_POLICY_AUTOMATIC,
	                                GTK_POLICY_AUTOMATIC);

	edb->tree = create_tree (edb);
	gtk_container_add (GTK_CONTAINER (scrolled),
	                   GTK_WIDGET (edb->tree));

	gtk_box_pack_start (GTK_BOX (vbox), scrolled,
	                    TRUE, TRUE, 6);
	gtk_box_pack_start (GTK_BOX (vbox), create_button_group (edb),
	                    FALSE, FALSE, 6);

	/* finally add the vbox to the notebook */
	edb->notebook_child = vbox;
	tab_header = create_notebook_page_header (edb, name,
	                                          &edb->tab_label);
	g_free (name);

	gtk_notebook_append_page (notebook, vbox, tab_header);
	gtk_widget_show_all (vbox);

	return edb;
}

void essence_db_free (EssenceDb *edb)
{
	int i = -1;

	assert (edb);

	if (edb->notebook_child)
		i = gtk_notebook_page_num (edb->notebook, edb->notebook_child);

	if (i != -1)
		gtk_notebook_remove_page (edb->notebook, i);

	while (edb->items) {
		essence_db_item_free (edb->items->data);
		edb->items = g_slist_remove_link (edb->items, edb->items);
	}

	e_db_close (edb->db);
	free (edb);
}

bool essence_db_load (EssenceDb *edb)
{
	EssenceDbItem *ei;
	char **keys;
	int i, num = 0;

	assert (edb);

	/* now read the Edb file */
	if (!(edb->db = e_db_open (edb->file)))
		return false;

	/* get a list of all the entries in the db */
	if (!(keys = e_db_dump_key_list (edb->file, &num)))
		return false;

	/* now read each DB entry and show it in the treeview */
	for (i = 0; i < num; i++) {
		if ((ei = essence_db_item_new (edb->db))) {
			essence_db_item_key_set (ei, keys[i]);

			if (essence_db_item_load (ei)) {
				edb->items = g_slist_append (edb->items, ei);
				essence_db_item_show (ei, edb->store);
			}
		}

		free (keys[i]);
	}

	free (keys);

	return true;
}

/**
 * Syncs the contents of the liststore with the EDb object.
 *
 * @param edb
 *
 * @return Boolean success or failure.
 */
bool essence_db_save (EssenceDb *edb)
{
	GSList *l;

	assert (edb);

	for (l = edb->items; l; l = l->next)
		essence_db_item_save (l->data);

	e_db_flush ();

	return true;
}

bool essence_db_save_as (EssenceDb *edb, const char *file)
{
	EssenceDbItem *ei;
	E_DB_File *new_db;
	GSList *l;
	char *name, markup[128];

	assert (edb);
	assert (file);

	/* if the old and the new name match, we just save it */
	if (!strcmp (file, edb->file))
		return essence_db_save (edb);

	/* try to open the new file read/write */
	if (!(new_db = e_db_open ((char *) file)))
		return false;

	/* success, so close the old one */
	e_db_close (edb->db);
	edb->db = new_db;
	strlcpy (edb->file, file, sizeof (edb->file));

	for (l = edb->items; l; l = l->next) {
		ei = l->data;
		ei->db = edb->db;
	}

	/* now adjust the tab header label to reflect the
	 * filename change
	 */
	name = g_path_get_basename ((char *) file);
	snprintf (markup, sizeof (markup),
	          "<span weight=\"bold\">%s</span>", name);
	gtk_label_set_markup (GTK_LABEL (edb->tab_label), markup);
	g_free (name);

	return essence_db_save (edb);
}
